#include                <stdlib.h>
#include                <dirent.h>
#include                <string.h>
#include                "tap.h"
#include                "libFireBird.h"

#define PROGRAM_NAME    "lost+found"
#define VERSION         "V2.0"

TAP_ID                  (0x8E0A4235);
TAP_PROGRAM_NAME        (PROGRAM_NAME" "VERSION);
TAP_PROGRAM_VERSION     (VERSION);
TAP_AUTHOR_NAME         ("FireBird");
TAP_DESCRIPTION         ("Move lost+found files into DataFiles");
TAP_ETCINFO             (__DATE__);

#define LOSTANDFOUNDDIR "/DataFiles/lost_and_found"
#define DBVERSION       1
#define DBDIR           "/ProgramFiles/Settings/lost_and_found"
#define DBFILE          "lost+found.db"
#define INIFILE         "lost+found.ini"
#define LOGFILE         "lost+found.log"

typedef struct
{
  __ino64_t             recInode;
  dword                 recSignature;           //4 Bytes, starting at byte 0x08
  dword                 unused1;
  __ino64_t             infInode;
  dword                 infSignature;           //4 Bytes, starting at byte 0x08 (usually the recording time)
  dword                 unused2;
  __ino64_t             navInode;
  dword                 navSignature;           //4 Bytes, starting at byte 0x0c
  dword                 unused3;
  char                  FileName[MAX_FILE_NAME_SIZE+1];
} tRecDatabase;

typedef struct
{
  __dev_t               st_dev;
  long int              st_pad1[2];
  __ino64_t             st_ino;
  __mode_t              st_mode;
  __nlink_t             st_nlink;
  __uid_t               st_uid;
  __gid_t               st_gid;
  __dev_t               st_rdev;
  long int              st_pad2[2];
  __off64_t             st_size;
  __time_t              st_atime;
  long int              __reserved0;
  __time_t              st_mtime;
  long int              __reserved1;
  __time_t              st_ctime;
  long int              __reserved2;
  __blksize_t           st_blksize;
  long int              st_pad4;
  __blkcnt64_t          st_blocks;
  long int              st_pad5[14];
} tstat64;

extern int lstat64(__const char *__restrict __file, tstat64 *__restrict __buf) __THROW;
void Settings_Write(void);

static tRecDatabase    *RecDB;
static int              RecDBVersion, RecDBSize, RecDBEntries;
static dword            NextLFCheck, NextDataFilesCheck, MessageTimeout;
static char             LogString[512];
static bool             InformUser = FALSE;

void WriteLog(char *s)
{
  HDD_TAP_PushDir();
  HDD_ChangeDir(DBDIR);
  LogEntry(LOGFILE, PROGRAM_NAME, TRUE, TIMESTAMP_YMDHMS, s);
  HDD_TAP_PopDir();
}

void Settings_Read(void)
{
  HDD_ChangeDir(DBDIR);

  if(INIOpenFile(INIFILE, PROGRAM_NAME) > INILOCATION_NewFile)
  {
    InformUser = INIGetInt("InformUser", 1, 0, 1) != 0;
  }
  else
  {
    //Default-Werte, wenn die INI nicht existiert
    InformUser = FALSE;

    Settings_Write();
  }

  INICloseFile();
}

void Settings_Write(void)
{
  HDD_ChangeDir(DBDIR);

  INIOpenFile(INIFILE, PROGRAM_NAME);
  INISetInt("InformUser", InformUser ? 1 : 0);

  INISaveFile(INIFILE, INILOCATION_AtCurrentDir, "");
  INICloseFile();
}

void RecDB_CreateDir(void)
{
  HDD_ChangeDir("/ProgramFiles");
  TAP_Hdd_Create("Settings", ATTR_FOLDER);
  HDD_ChangeDir("Settings");
  TAP_Hdd_Create("lost_and_found", ATTR_FOLDER);
}

void RecDB_Load(void)
{
  TYPE_File            *f;

  HDD_TAP_PushDir();

  HDD_ChangeDir(DBDIR);
  if(TAP_Hdd_Exist(DBFILE))
  {
    f = TAP_Hdd_Fopen(DBFILE);
    if(f)
    {
      if(RecDB) TAP_MemFree(RecDB);
      TAP_Hdd_Fread(&RecDBVersion, 1, sizeof(int), f);
      TAP_Hdd_Fread(&RecDBSize, 1, sizeof(int), f);
      TAP_Hdd_Fread(&RecDBEntries, 1, sizeof(int), f);

      if(RecDBVersion == DBVERSION)
      {
        RecDB = TAP_MemAlloc(RecDBSize * sizeof(tRecDatabase));
        if(RecDB)
        {
          TAP_Hdd_Fread(RecDB, RecDBSize, sizeof(tRecDatabase), f);
        }
        else
        {
          WriteLog("failed to reserve memory for the DB");
          RecDBSize = 0;
          RecDBEntries = 0;
        }
      }
      else
      {
        WriteLog("unexpected DB version");
        RecDBSize = 0;
        RecDBEntries = 0;
      }

      TAP_Hdd_Fclose(f);
    }
    else
    {
      WriteLog("failed to load the DB");
    }
  }

  HDD_TAP_PopDir();
}

void RecDB_Save(void)
{
  TYPE_File            *f;

  if(RecDB)
  {
    HDD_TAP_PushDir();

    HDD_ChangeDir(DBDIR);
    TAP_Hdd_Delete(DBFILE);
    TAP_Hdd_Create(DBFILE, ATTR_NORMAL);
    f = TAP_Hdd_Fopen(DBFILE);
    if(f)
    {
      RecDBVersion = DBVERSION;
      TAP_Hdd_Fwrite(&RecDBVersion, 1, sizeof(int), f);
      TAP_Hdd_Fwrite(&RecDBSize, 1, sizeof(int), f);
      TAP_Hdd_Fwrite(&RecDBEntries, 1, sizeof(int), f);
      TAP_Hdd_Fwrite(RecDB, RecDBSize, sizeof(tRecDatabase), f);
      TAP_Hdd_Fclose(f);
    }
    else
    {
      WriteLog("failed to save the DB");
    }
    HDD_TAP_PopDir();
  }
}

void RecDB_Destroy(void)
{
  RecDBSize = 0;
  RecDBEntries = 0;
  if(RecDB) TAP_MemFree(RecDB);
  RecDB = NULL;
}

bool RecDB_Add(tRecDatabase *RecFile)
{
  int                   i;
  tRecDatabase         *DB;
  bool                  ret;

  ret = FALSE;

  DB = RecDB;
  for(i = 0; i < RecDBEntries; i++)
  {
    if((RecFile->recInode == DB->recInode) && (RecFile->recSignature == DB->recSignature))
    {
      //Record found, check if there are changes
      if(DB->infInode != RecFile->infInode)
      {
        DB->infInode = RecFile->infInode;
        ret = TRUE;
      }

      if(DB->infSignature != RecFile->infSignature)
      {
        DB->infSignature = RecFile->infSignature;
        ret = TRUE;
      }

      if(DB->navInode != RecFile->navInode)
      {
        DB->navInode = RecFile->navInode;
        ret = TRUE;
      }

      if(DB->navSignature != RecFile->navSignature)
      {
        DB->navSignature = RecFile->navSignature;
        ret = TRUE;
      }

      if(strcmp(DB->FileName, RecFile->FileName))
      {
        strcmp(DB->FileName, RecFile->FileName);
        ret = TRUE;
      }

      return ret;
    }
    DB++;
  }

  //No record found, add it
  //More space needed?
  if(RecDBEntries >= RecDBSize)
  {
    tRecDatabase       *tempRecDB;

    tempRecDB = TAP_MemAlloc((RecDBSize + 10) * sizeof(tRecDatabase));
    if(tempRecDB == NULL)
    {
      WriteLog("failed to reserve memory while increasing the DB");
      return FALSE;
    }

    memset(tempRecDB, 0, (RecDBSize + 10) * sizeof(tRecDatabase));
    if(RecDB)
    {
      memcpy(tempRecDB, RecDB, RecDBEntries * sizeof(tRecDatabase));
      TAP_MemFree(RecDB);
    }
    RecDB = tempRecDB;
    RecDBSize += 10;
  }

  memcpy(&RecDB[RecDBEntries], RecFile, sizeof(tRecDatabase));
  RecDBEntries++;

  TAP_SPrint(LogString, "added '%s' to the DB", RecFile->FileName);
  WriteLog(LogString);

  return TRUE;
}

bool isNewLostAndFound(void)
{
  DIR                *dirp;
  struct dirent      *dp;
  int                 Entries = 0;

  dirp = opendir("/mnt/hd/lost+found");
  if(dirp)
  {
    while((dp = readdir(dirp)) != NULL)
      Entries++;

    closedir(dirp);
  }

  if(Entries > 2)
  {
    TAP_SPrint(LogString, "found %d new files in the lost+found folder", Entries - 2);
    WriteLog(LogString);
  }

  return (Entries > 2);
}

void MoveAllLostAndFoundFiles(void)
{
  system("mkdir /mnt/hd" LOSTANDFOUNDDIR);
  system("mv /mnt/hd/lost+found/* /mnt/hd" LOSTANDFOUNDDIR);
}

bool CrossCheckLostAndFoundWithDB(void)
{
  int                   NrFiles, i, j;
  TYPE_FolderEntry      FolderEntry;
  TYPE_File            *f;
  tstat64               statbuf;
  dword                 Sig08, Sig0C;
  tRecDatabase         *DB;
  bool                  ret;

  char                  AbsFileName[512];

  ret = FALSE;

  HDD_ChangeDir(LOSTANDFOUNDDIR);
  NrFiles = TAP_Hdd_FindFirst(&FolderEntry, "rcn");
  for (i = 0; i < NrFiles; i++)
  {
    if(FolderEntry.attr == ATTR_NORMAL)
    {
      TAP_SPrint(AbsFileName, "%s%s/%s", TAPFSROOT, LOSTANDFOUNDDIR, FolderEntry.name);
      if(lstat64(AbsFileName, &statbuf) == 0)
      {
        f = TAP_Hdd_Fopen(FolderEntry.name);
        if(f)
        {
          TAP_Hdd_Fseek(f, 8, SEEK_SET);
          TAP_Hdd_Fread(&Sig08, 1, sizeof(dword), f);
          TAP_Hdd_Fread(&Sig0C, 1, sizeof(dword), f);
          TAP_Hdd_Fclose(f);

          DB = RecDB;
          for(j = 0; j < RecDBEntries; j++)
          {
            if((DB->recInode == statbuf.st_ino) && (DB->recSignature == Sig08))
            {
              TAP_Hdd_Rename(FolderEntry.name, DB->FileName);

              TAP_SPrint(LogString, "recovered '%s' to '%s'", FolderEntry.name, DB->FileName);
              WriteLog(LogString);
              ret = TRUE;
              break;
            }

            if((DB->infInode == statbuf.st_ino) && (DB->infSignature == Sig08))
            {
              TAP_SPrint(AbsFileName, "%s.inf", DB->FileName);
              TAP_Hdd_Rename(FolderEntry.name, AbsFileName);

              TAP_SPrint(LogString, "recovered '%s' to '%s'", FolderEntry.name, AbsFileName);
              WriteLog(LogString);
              ret = TRUE;
              break;
            }

            if((DB->navInode == statbuf.st_ino) && (DB->navSignature == Sig0C))
            {
              TAP_SPrint(AbsFileName, "%s.nav", DB->FileName);
              TAP_Hdd_Rename(FolderEntry.name, AbsFileName);

              TAP_SPrint(LogString, "recovered '%s' to '%s'", FolderEntry.name, AbsFileName);
              WriteLog(LogString);
              ret = TRUE;
              break;
            }

            DB++;
          }
        }
        else
        {
          TAP_SPrint(LogString, "failed to open '%s' while recovering", FolderEntry.name);
          WriteLog(LogString);
        }
      }
    }
    TAP_Hdd_FindNext(&FolderEntry);
  }

  return ret;
}

bool CrossCheckDataFilesWithDB(void)
{
  tStringDB             StringDB;
  int                   NrFiles, i, MaxLoop;
  TYPE_FolderEntry      FolderEntry;
  TYPE_File            *f;
  tstat64               statbuf;
  char                  AbsFileName[512], NextDir[512];
  tRecDatabase          RecFile;
  bool                  ret;
  static bool           FirstCall = TRUE;

  ret = FALSE;

  MaxLoop = 0;
  StringDBInit(&StringDB, 1024);
  StringDBAdd(&StringDB, "/DataFiles");

  while(!StringDBEOF(&StringDB))
  {
    if(MaxLoop++ > 1000)
    {
      WriteLog("endless loop exit while scanning");
      FirstCall = FALSE;
      return ret;
    }

    HDD_ChangeDir(StringDBGet(&StringDB, StringDBCurrent(&StringDB)));

    NrFiles = TAP_Hdd_FindFirst(&FolderEntry, "rec|mpg");
    for (i = 0; i < NrFiles; i++)
    {
      //Do a complete recursive scan on the first call
      if(FirstCall && (FolderEntry.attr == ATTR_FOLDER))
      {
        TAP_SPrint(NextDir, "%s/%s", StringDBGet(&StringDB, StringDBCurrent(&StringDB)), FolderEntry.name);

        //But only if it is not mounted to an external device
        TAP_SPrint(AbsFileName, "%s%s", TAPFSROOT, NextDir);
        if((lstat64(AbsFileName, &statbuf) == 0) && (statbuf.st_dev == 0x802)) StringDBAdd(&StringDB, NextDir);
      }

      if(FolderEntry.attr == ATTR_NORMAL)
      {
        TAP_SPrint(AbsFileName, "%s%s/%s", TAPFSROOT, StringDBGet(&StringDB, StringDBCurrent(&StringDB)), FolderEntry.name);
        if(lstat64(AbsFileName, &statbuf) == 0)
        {
          memset(&RecFile, 0, sizeof(tRecDatabase));

          RecFile.recInode = statbuf.st_ino;
          strcpy(RecFile.FileName, FolderEntry.name);

          f = TAP_Hdd_Fopen(FolderEntry.name);
          if(f)
          {
            TAP_Hdd_Fseek(f, 0x08, SEEK_SET);
            TAP_Hdd_Fread(&RecFile.recSignature, 1, sizeof(dword), f);
            TAP_Hdd_Fclose(f);

            //Check for the inf
            TAP_SPrint(AbsFileName, "%s%s/%s.inf", TAPFSROOT, StringDBGet(&StringDB, StringDBCurrent(&StringDB)), FolderEntry.name);
            if(lstat64(AbsFileName, &statbuf) == 0)
            {
              RecFile.infInode = statbuf.st_ino;

              TAP_SPrint(AbsFileName, "%s.inf", FolderEntry.name);
              f = TAP_Hdd_Fopen(AbsFileName);
              if(f)
              {
                TAP_Hdd_Fseek(f, 0x08, SEEK_SET);
                TAP_Hdd_Fread(&RecFile.infSignature, 1, sizeof(dword), f);
                TAP_Hdd_Fclose(f);
              }
              else
              {
                TAP_SPrint(LogString, "failed to open '%s.inf' while scanning", FolderEntry.name);
                WriteLog(LogString);
              }
            }

            //Check for the nav
            TAP_SPrint(AbsFileName, "%s%/%s.nav", TAPFSROOT, StringDBGet(&StringDB, StringDBCurrent(&StringDB)), FolderEntry.name);
            if(lstat64(AbsFileName, &statbuf) == 0)
            {
              RecFile.navInode = statbuf.st_ino;
              TAP_SPrint(AbsFileName, "%s.nav", FolderEntry.name);
              f = TAP_Hdd_Fopen(AbsFileName);
              if(f)
              {
                TAP_Hdd_Fseek(f, 0x0c, SEEK_SET);
                TAP_Hdd_Fread(&RecFile.navSignature, 1, sizeof(dword), f);
                TAP_Hdd_Fclose(f);
              }
              else
              {
                TAP_SPrint(LogString, "failed to open '%s.nav' while scanning", FolderEntry.name);
                WriteLog(LogString);
              }
            }

            ret = ret | RecDB_Add(&RecFile);
          }
          else
          {
            TAP_SPrint(LogString, "failed to open '%s' while scanning", FolderEntry.name);
            WriteLog(LogString);
          }
        }
        else
        {
          TAP_SPrint(LogString, "failed to stat '%s' while scanning", AbsFileName);
          WriteLog(LogString);
        }
      }
      TAP_Hdd_FindNext(&FolderEntry);
    }
    StringDBNext(&StringDB);
  }
  StringDBDestroy(&StringDB);

  FirstCall = FALSE;
  return ret;
}

dword TAP_EventHandler(word event, dword param1, dword param2)
{
  (void) param2;

  if(event == EVT_KEY)
  {
    if(((param1 == RKEY_Ok) || (param1 == RKEY_Exit)) && OSDMenuMessageBoxIsVisible())
    {
      WriteLog("Dialog ack'd by user");
      InformUser = FALSE;
      Settings_Write();
      OSDMenuMessageBoxDestroy();
      return 0;
    }
  }

  if(event == EVT_IDLE)
  {
    if(MessageTimeout && (TAP_GetTick() > MessageTimeout))
    {

      OSDMenuMessageBoxDestroy();
      MessageTimeout = 0;
    }

    if(TAP_GetTick() > NextDataFilesCheck)
    {
      RecDB_Load();
      if(CrossCheckDataFilesWithDB()) RecDB_Save();
      RecDB_Destroy();
      NextDataFilesCheck = TAP_GetTick() + 30000;
    }

    if(TAP_GetTick() > NextLFCheck)
    {
      if(isNewLostAndFound())
      {
        InformUser = TRUE;
        Settings_Write();

        MoveAllLostAndFoundFiles();
        RecDB_Load();
        CrossCheckLostAndFoundWithDB();
        RecDB_Destroy();
      }
      NextLFCheck = 0xffffffff;

      if(InformUser)
      {
        OSDMenuMessageBoxInitialize(PROGRAM_NAME, "Dateien gefunden und verschoben");
        OSDMenuMessageBoxButtonAdd("OK");
        OSDMenuMessageBoxButtonSelect(0);
        OSDMenuMessageBoxShow();
        MessageTimeout = TAP_GetTick() + 3000;
      }
    }
  }

  return param1;
}

int TAP_Main(void)
{
  RecDB = NULL;
  RecDBSize = 0;
  RecDBEntries = 0;

  if(GetUptime() > 12000)
  {
    NextLFCheck = TAP_GetTick() + 100;            //1 second
    NextDataFilesCheck = TAP_GetTick() + 4000;    //40 seconds
  }
  else
  {
    NextLFCheck = TAP_GetTick() + 18000;          //3 minutes
    NextDataFilesCheck = TAP_GetTick() + 12000;   //2 minutes
  }

  MessageTimeout = 0;

  RecDB_CreateDir();
  Settings_Read();

  return 1;
}
